structure Check :>
sig

    exception TypeError of string

    type memory_typing = (Syntax.location * Syntax.typ) list

    (* Given a memory typing Sigma and an expression e, 
       return tau if there is a tau such that Sigma | . |- e : tau, or
       raise TypeError otherwise *)
    val check : memory_typing * Syntax.exp -> Syntax.typ

    (* Given a memory mu, return () if mu is well-typed or raise TypeError otherwise *)
    val check_memory : Syntax.memory -> memory_typing

    val check_state : Syntax.memory * Syntax.exp -> Syntax.typ 
        
end  = 
struct

    open Syntax
    
    type context = (name * typ) list

    type memory_typing = (location * typ) list
        
    exception TypeError of string

    fun lookup (gamma : context, v : name) : typ = 
	case List.find (fn (v', t) => v = v') gamma of
            NONE => raise TypeError ("unbound variable " ^ nameToString v)
          | SOME (_, t) => t

    fun lookupLoc (s : memory_typing, l : location) : typ = 
	case List.find (fn (l', t) => l = l') s of
            NONE => raise TypeError ("unbound location " ^ locToString l)
          | SOME (_, t) => t

    (* Given a context gamma and expression e, 
       return tau if there is a tau such that gamma |- e : tau, or
       raise TypeError otherwise *)
    fun checkOpen (s : memory_typing, g : context, e : exp) : typ  = 
        case freshenTop e of 
            Var v => lookup (g,v)
          | Loc l => lookupLoc(s,l)
          | Lam (t,(x,e)) => Arrow (t,checkOpen (s, (x,t) :: g, e))
          | App (e1,e2) => 
                (case checkOpen (s, g, e1) of
                     Arrow (t,t') => let val _ = checkWithExpected (s, g, e2, t) in t' end
                   | _ => raise TypeError "function position not of function type")
          | Let (e1, t, (x,e2)) => checkOpen (s, ((x,checkWithExpected (s, g, e1, t)) :: g), e2 )
          | Zero => Nat
          | Succ e => checkWithExpected(s,g,e,Nat)
          | NatCase (e, e0, (x,e1)) =>
                            let
                                val _ = checkWithExpected(s, g,e,Nat)
                                val t = checkOpen(s, g,e0)
                                val _ = checkWithExpected(s, (x,Nat) :: g, e1, t)
                            in
                                t
                            end
          | Nil => List
          | Cons(e1, e2) => let val _ = checkWithExpected(s, g,e1,Nat)
                                val _ = checkWithExpected(s, g,e2,List)
                            in 
                                List
                            end
          | ListCase (e, e0, (x,y,e1)) =>
                            let
                                val _ = checkWithExpected(s, g,e,List)
                                val t = checkOpen(s, g,e0)
                                val _ = checkWithExpected(s, (x,Nat) :: (y,List) :: g, e1, t)
                            in
                                t
                            end
          | Fix(t,(x,e)) => checkWithExpected(s, (x,t) :: g, e, t)

    and checkWithExpected(s : memory_typing, g : context, e : exp, t : typ) : typ =
        if checkOpen(s, g,e) = t then t else raise TypeError "expression doesn't have expected type"
                     
    (* Given an expression e, 
       return tau if there is a tau such that empty |- e : tau,
       or raise TypeError if e is not well-typed in the empty context *)
    fun check (s,e) = checkOpen (s, [], e)

    fun check_memory(m) = let val S = List.map (fn (l,t,_) => (l,t)) m
                              val _ = List.map (fn (l,t,e) => case e of BlackHole => t | Defined e => checkWithExpected(S,[],e,t)) m
                          in
                              S
                          end 

    fun check_state(m,e) = let val S = check_memory m
                           in check(S,e)
                           end 
                      
end
