signature PARSE =
sig

  val parse : (Lexer.token * char list) Stream.stream -> Syntax.exp Stream.stream

  exception Error of string
end;  (* signature PARSE *)

structure Parse :> PARSE = 
struct

    type exp = Syntax.exp
    
    structure S = Stream
    open Syntax
    open Lexer

    exception Error of string

    (* 
       The concrete grammar parsed by this module is as follows:

       program ::= exp0 SEMICOLON

       typ ::= NAT | LPAREN typ ARR typ RPAREN | LPAREN typ PRODUCT typ RPAREN | LPAREN typ SUM typ RPAREN | rec VAR(s) DOT type
       
       exp0    ::= exp0 APP expinf | expinf
       expinf  ::= LPAREN exp0 RPAREN
                |  NUMBER(k)
                |  VAR(s)
                |  LET VAR(s) BE exp0 IN exp0 END
                |  ZERO
                |  SUCC exp0 
                |  REC exp0 LCURLY ZERO GOESTO exp0 BAR SUCC VAR(s) WITH VAR(s) GOESTO exp0 RCURLY
                |  LAM var(S) COLON typ DOT exp0
                |  LANGLE exp0 COMMA exp0 RANGLE 
                |  LANGLE RANGLE 
                |  INLEFT LBRACKET typ RBRACKET exp0 
                |  INRIGHT LBRACKET typ RBRACKET exp0 
                |  CASE exp0 LCURLY INLEFT VAR(s) GOESTO exp0 BAR INRIGHT VAR(s) GOESTO exp0 RCURLY

    *)

    exception Unbound of string

    fun Num 0 = Zero
      | Num n = Succ ( (Num (n-1)))

    (* next s = (x,s'), where x is the head of s, s' the tail of s
       raises Error if stream is empty
    *)
    fun next s =
        case S.force s
          of S.Nil => raise Error "Unexpected end of stream"
           | S.Cons result => result

    type pstream = (Lexer.token * char list) Stream.stream 
    type env = string -> name
        
    fun expected (s : pstream) (msg : string) : 'a =
        let val ((_, l), _) = next s
        in raise Error ("Expected " ^ msg ^ " at `" ^
                        concat (map Char.toString l) ^ "'") end

    fun require (s : pstream) (token : token) =
            let val ((t,l), s') = next s
            in if t = token then s'
               else expected s (Lexer.tokenToString token ^ " token")
            end

    (* NOTE: I left in the inputs for parsing type variables from the code this was copied from,
             but they aren't currently used for this assignment.  
     *) 
        
    (* d is type environment 
       g is expression environment *)

    fun extend (g : env, x : string) : env =
        let
            val v = Syntax.freshNameString x in (fn y => if x = y then v else g y)
        end
    
    fun var (s : pstream) : string * pstream =
        case next s of
            ((VAR(x),_), s) => (x, s)
          | _ => expected s "a variable name"
                
    fun program (s : pstream) : exp * pstream =
        let
            val (e, s) = exp0 ((fn x => raise Error ("Free type variable `" ^ x ^ "'")),
                               (fn x => raise Error ("Free variable `" ^ x ^ "'")),
                               s) 
                val s = require s SEMICOLON
        in (e, s)
        end

    and typ (d : env, s : pstream) : typ * pstream = 
        case next s of
            ((NAT,_),s) => (Syntax.Nat,s)
          | ((LIST,_),s) => (Syntax.List,s)
          | ((LPAREN,_),s) => let val (t1,s) = typ (d,s)
                                  val ((middlet,_),s) = next s
                                  val con = (case middlet of
                                                 ARR => Syntax.Arrow
                                               | _ => expected s "an infix type constructor") 
                                  val (t2,s) = typ (d,s)
                                  val s = require s RPAREN
                              in
                                  (con(t1,t2),s)
                              end
          | ((t,_),s) => expected s "a type" 

    and exp0 (d : env, g : env, s : pstream) : exp * pstream =
        let
            val (e, s) = expinf (d, g, s)
        in
            exp0' (e, d, g, s)
        end
    
   and exp0' (e : exp, d : env, g : env, s : pstream) : exp * pstream =
       case next s
           of ((APP,_), s) => 
               let val (e', s) = expinf (d, g, s)
               in exp0' ( (App(e, e')), d, g, s) end
         | _ => (e, s)
                   
   and expinf (d : env, g : env, s : pstream) : exp * pstream =
       case next s of
           ((LPAREN,_), s) => let val (e, s) = exp0 (d, g, s)
                                  val s = require s RPAREN
                              in (e, s) end
         | ((NUMBER(k),_), s) => ( (Num(k)), s)
         | ((VAR(x),_), s) => ( (Var(g x)), s)
         | ((LET,_), s) => let val (x, s) = var s
                               val s = require s COLON
                               val (t,s) = typ (d,s) 
                               val s = require s BE
                               val (e1, s) = exp0 (d, g, s)
                               val s = require s IN
                               val g' = extend(g, x)
                               val (e2, s) = exp0 (d, g', s)
                               val s = require s END
                           in ( (Let(e1, t, (g' x, e2))), s) end
         | ((ZERO,_) , s) => ( Zero , s)
         | ((SUCC,_) , s) => let val (e,s) = exp0 (d, g , s) in ( (Succ e) , s) end
         | ((NIL,_) , s) => ( Nil , s)
         | ((CONS,_) , s) => let val s = require s LEFTANGLE 
                                 val (e1,s) = exp0 (d, g , s)
                                 val s = require s COMMA
                                 val (e2,s) = exp0 (d, g , s)
                                 val s = require s RIGHTANGLE
                             in 
                                 (Cons(e1,e2) , s)
                             end
         | ((LAM,_),s) => let val (x, s) = var s
                              val s = require s COLON
                              val (t,s) = typ (d,s)
                              val s = require s DOT
                              val g' = extend (g,x)
                              val (e,s) = exp0 (d, g',s)
                          in
                              ( (Lam(t,(g' x,e))),s)
                          end
         | ((NATCASE,_),s) => let val (e,s) = exp0 (d, g,s)
                                  val s = require s LEFTCURLY
                                  val s = require s ZERO
                                  val s = require s GOESTO
                                  val (e0,s) = exp0 (d, g,s)
                                  val s = require s BAR
                                  val s = require s SUCC
                                  val (x,s) = var s
                                  val s = require s GOESTO
                                  val g' = extend(g,x)
                                  val (e1,s) = exp0 (d, g',s)
                                  val s = require s RIGHTCURLY
                              in 
                                  ( (NatCase(e,e0,(g' x, e1))), s)
                              end
          | ((LISTCASE,_),s) => let val (e,s) = exp0 (d, g,s)
                                    val s = require s LEFTCURLY
                                    val s = require s NIL
                                    val s = require s GOESTO
                                    val (e0,s) = exp0 (d, g,s)
                                    val s = require s BAR
                                    val s = require s CONS
                                    val s = require s LEFTANGLE 
                                    val (x,s) = var s
                                    val s = require s COMMA
                                    val (y,s) = var s
                                    val s = require s RIGHTANGLE
                                    val s = require s GOESTO
                                    val g' = extend(extend(g,x),y)
                                    val (e1,s) = exp0 (d, g',s)
                                    val s = require s RIGHTCURLY
                                in 
                               ( (ListCase(e,e0,(g' x, g' y, e1))), s)
                           end
         | ((FIX,_),s) => let val (x, s) = var s
                              val s = require s COLON
                              val (t,s) = typ (d,s)
                              val s = require s DOT
                              val g' = extend (g,x)
                              val (e,s) = exp0 (d, g',s)
                          in
                              (Fix(t,(g' x,e)),s)
                          end
         | _ => expected s "an expression"

        
    (* Exported parsing function *)
          
    fun parse s = S.delay (fn () =>
            case S.force s
              of S.Nil => S.Nil
               | S.Cons _ => let val (e, s) = program s
                             in S.Cons(e, parse s) end)

end;  (* functor ParseFun *)
