structure Syntax :>
sig

    datatype typ = Nat | Arrow of typ * typ | List;

    (* for this assignment, we represent names by a pair of a string and
       a unique int suffix (used for alpha-conversion).  
       freshName generates a new suffix that has never been used before,
       so freshName x is guaranteed to be different from all previously used variables.  

       Note that you can still use (x = y) to check if two names are equal.  

       In a term, variable names (x,i) are printed as x[i].  You can
       mostly ignore [i] when reading the term, and only look at them
       for disambiguation if things seem wrong.  

       *)
    type name = string * int
    val freshNameString : string -> name
    val freshName : name -> name
    val nameToString : name -> string

    type location = int
    val freshLoc : unit -> location
    val locToString : location -> string
        
    datatype exp = 
	Var of name
      | Loc of location (* location, written @ a in PFPL *)
      | Zero
      | Succ of exp
      | NatCase of exp * exp * (name * exp) 
      | Nil
      | Cons of exp * exp
      | ListCase of exp * exp * (name * name * exp) 
      | Lam of typ * (name * exp)
      | App of exp * exp
      | Let of exp * typ * (name * exp)
      | Fix of typ * (name * exp)

    (* freshen the top-level bound variables in e if e is a Let/Lam/Rec *)
    val freshenTop : exp -> exp
        
    val subst : exp * (exp * name) -> exp

    datatype memory_entry = Defined of exp | BlackHole 
        
    type memory = (location * typ * memory_entry) list

end =
struct

    datatype typ = Nat | Arrow of typ * typ | List;

    type name = string * int

    local 
	val suffix = ref 0 
    in
        fun freshNameString (s : string) : name = 
            (suffix := (!suffix)+1;
             (s,(!suffix)))
            
	fun freshName ((s,i) : name) : name = 
	    freshNameString s
    end

    fun nameToString (x,s) = x ^ "[" ^ Int.toString s ^"]"

    type location = int
    local 
	val newloc = ref 0 
    in
        fun freshLoc () : int = 
            (newloc := (!newloc)+1;
             (!newloc))
        
    end
    fun locToString (x) = "l[" ^ Int.toString x ^"]"

    datatype exp = 
	Var of name
      | Loc of location
      | Zero
      | Succ of exp
      | NatCase of exp * exp * (name * exp) 
      | Nil
      | Cons of exp * exp
      | ListCase of exp * exp * (name * name * exp) 
      | Lam of typ * (name * exp)
      | App of exp * exp
      | Let of exp * typ * (name * exp)
      | Fix of typ * (name * exp)

    fun swapName (z : name, (x : name, y : name)) =
	if z = x then y
	else if z = y then x
	else z

    fun swapExp (e : exp, (z : name, w : name)) : exp =
	let (* give short names to the recursive call and 
               swapName call since they always get called with the same x and y *)
            fun se e = swapExp (e, (z, w))
	    fun sn x = swapName (x ,(z, w))
        in
            case e of 
                Var z => Var(sn z)
              | Loc l => Loc l
              | Zero => Zero
              | Succ e => Succ (se e)
              | NatCase (e, e0, (x,e1)) => NatCase (se e, se e0 , (sn x, se e1))
              | Lam (t, (x,e)) => Lam (t, (sn x, se e))
              | App (e1, e2) => App (se e1 , se e2)
              | Let(e1, t, (x, e2)) => Let(se e1, t, (sn x, se e2))
              | Fix(t,(x,e)) => Fix (t, (sn x, se e))
              | Nil => Nil
              | Cons (e1,e2) => Cons (se e1, se e2)
              | ListCase (e, e0, (x,y,e1)) => ListCase (se e, se e0 , (sn x,sn y, se e1))
	end


    (* freshen the topmost binder in e *)
    fun freshenTop (e : exp) : exp =
        case e of 
            (Let (e1, t, (x, e2))) => 
                let val x' = freshName x in
                    Let (e1, t, (x', swapExp (e2, (x, x'))))
                end
          | (NatCase (e, e0, (x,e1))) =>
                let val x' = freshName x
                in NatCase (e, e0 , (x',swapExp(e1, (x,x'))))
                end
          | (Lam (t, (x,e))) => 
                let val x' = freshName x
                in Lam (t,(x',swapExp(e, (x,x'))))
                end
          | (ListCase (e, e0, (x,y,e1))) =>
                let val x' = freshName x
                    val y' = freshName y
                in ListCase (e, e0 , (x',y', swapExp(swapExp (e1, (x,x')), (y,y'))))
                end
          | Fix (t, (x,e)) => 
                let val x' = freshName x
                in Fix (t,(x',swapExp(e, (x,x'))))
                end
          | e => e

    fun subst (e : exp, (e' : exp , x' : name)) = 
	let (* give short names to the recursive call since they always get called with the same x and y *)
            fun se e = subst (e, (e', x'))
        in
            (* since we freshen the outermost bound variable here,
               the substitution will work even with shadowing and not capture
               *)
            case freshenTop e of
                Var x => if x = x' then e' else Var x
              | Loc l => Loc l 
              | Zero => Zero
              | Succ e => Succ (se e)
              | NatCase (e, e0, (x,e1)) => NatCase (se e, se e0 , (x, se e1))
              | Lam (t, (x,e)) => Lam (t, (x, se e))
              | App (e1, e2) => App (se e1 , se e2)
              | Let(e1, t, (x, e2)) => Let (se e1, t, (x, se e2))
              | Nil => Nil
              | Cons (e1,e2) => Cons (se e1 , se e2)
              | ListCase (e, e0, (x,y,e1)) => ListCase (se e, se e0 , (x,y, se e1))
              | Fix(t,(x,e)) => Fix(t,(x,se e))
        end

    datatype memory_entry = Defined of exp | BlackHole 
        
    type memory = (location * typ * memory_entry) list
    
end